==> symbol-processing-aa-embeddable-2.1.0-1.0.29-sources.jar: org/jetbrains/kotlin/ir/backend/js/utils/IrJsUtils.kt <==
/*
 * Copyright 2010-2021 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.ir.backend.js.utils

import ksp.org.jetbrains.kotlin.ir.util.parents
import ksp.org.jetbrains.kotlin.descriptors.isClass
import ksp.org.jetbrains.kotlin.descriptors.isInterface
import ksp.org.jetbrains.kotlin.ir.UNDEFINED_OFFSET
import ksp.org.jetbrains.kotlin.ir.backend.js.JsIrBackendContext
import ksp.org.jetbrains.kotlin.ir.backend.js.JsLoweredDeclarationOrigin
import ksp.org.jetbrains.kotlin.ir.backend.js.JsStatementOrigins
import ksp.org.jetbrains.kotlin.ir.backend.js.constructorFactory
import ksp.org.jetbrains.kotlin.ir.backend.js.defaultConstructorForReflection
import ksp.org.jetbrains.kotlin.ir.backend.js.export.isExported
import ksp.org.jetbrains.kotlin.ir.backend.js.ir.JsIrBuilder
import ksp.org.jetbrains.kotlin.ir.backend.js.lower.*
import ksp.org.jetbrains.kotlin.ir.declarations.*
import ksp.org.jetbrains.kotlin.ir.expressions.IrExpression
import ksp.org.jetbrains.kotlin.ir.expressions.IrFunctionAccessExpression
import ksp.org.jetbrains.kotlin.ir.expressions.IrInlinedFunctionBlock
import ksp.org.jetbrains.kotlin.ir.expressions.IrReturn
import ksp.org.jetbrains.kotlin.ir.expressions.impl.IrConstImpl
import ksp.org.jetbrains.kotlin.ir.expressions.impl.IrGetFieldImpl
import ksp.org.jetbrains.kotlin.ir.symbols.IrFunctionSymbol
import ksp.org.jetbrains.kotlin.ir.symbols.IrReturnableBlockSymbol
import ksp.org.jetbrains.kotlin.ir.types.IrType
import ksp.org.jetbrains.kotlin.ir.types.classifierOrNull
import ksp.org.jetbrains.kotlin.ir.util.*
import ksp.org.jetbrains.kotlin.name.FqName

val IrFile.nameWithoutExtension: String get() = name.substringBeforeLast(".kt")

fun IrClass.jsConstructorReference(context: JsIrBackendContext): IrExpression {
    return JsIrBuilder.buildCall(context.intrinsics.jsClass, origin = JsStatementOrigins.CLASS_REFERENCE)
        .apply { putTypeArgument(0, defaultType) }
}

fun IrDeclaration.isExportedMember(context: JsIrBackendContext) =
    parentClassOrNull != null && isExported(context)

fun IrDeclaration?.isExportedClass(context: JsIrBackendContext) =
    this is IrClass && kind.isClass && isExported(context)

fun IrDeclaration?.isExportedInterface(context: JsIrBackendContext) =
    this is IrClass && kind.isInterface && isExported(context)

fun IrReturn.isTheLastReturnStatementIn(target: IrReturnableBlockSymbol): Boolean {
    val ownerFirstStatement = target.owner.statements.singleOrNull()
    if (ownerFirstStatement is IrInlinedFunctionBlock) {
        return ownerFirstStatement.statements.lastOrNull() === this
    }
    return target.owner.statements.lastOrNull() === this
}

fun IrDeclarationWithName.getFqNameWithJsNameWhenAvailable(shouldIncludePackage: Boolean): FqName {
    val name = getJsNameOrKotlinName()
    return when (val parent = parent) {
        is IrDeclarationWithName -> parent.getFqNameWithJsNameWhenAvailable(shouldIncludePackage).child(name)
        is IrPackageFragment -> getKotlinOrJsQualifier(parent, shouldIncludePackage)?.child(name) ?: FqName(name.identifier)
        else -> FqName(name.identifier)
    }
}

fun IrConstructor.hasStrictSignature(context: JsIrBackendContext): Boolean {
    val primitives = with(context.irBuiltIns) { primitiveTypesToPrimitiveArrays.values + stringClass }
    return with(parentAsClass) {
        isExternal || isExpect || isAnnotationClass || context.inlineClassesUtils.isClassInlineLike(this) || symbol in primitives
    }
}

private fun getKotlinOrJsQualifier(parent: IrPackageFragment, shouldIncludePackage: Boolean): FqName? {
    return (parent as? IrFile)?.getJsQualifier()?.let { FqName(it) } ?: parent.packageFqName.takeIf { shouldIncludePackage }
}

val IrFunctionAccessExpression.typeArguments: List<IrType?>
    get() = List(typeArgumentsCount) { getTypeArgument(it) }

val IrFunctionAccessExpression.valueArguments: List<IrExpression?>
    get() = List(valueArgumentsCount) { getValueArgument(it) }

val IrClass.isInstantiableEnum: Boolean
    get() = isEnumClass && !isExpect && !isEffectivelyExternal()

val IrDeclaration.parentEnumClassOrNull: IrClass?
    get() = parents.filterIsInstance<IrClass>().firstOrNull { it.isInstantiableEnum }

fun IrFunctionSymbol.isUnitInstanceFunction(context: JsIrBackendContext): Boolean {
    return owner.origin === JsLoweredDeclarationOrigin.OBJECT_GET_INSTANCE_FUNCTION &&
            owner.returnType.classifierOrNull === context.irBuiltIns.unitClass
}

// TODO: the code is written to pass Repl tests, so we should understand. why in Repl tests we don't have backingField
fun JsIrBackendContext.getVoid(): IrExpression =
    intrinsics.void.owner.backingField?.let {
        IrGetFieldImpl(
            UNDEFINED_OFFSET,
            UNDEFINED_OFFSET,
            it.symbol,
            irBuiltIns.nothingNType
        )
    } ?: IrConstImpl.constNull(
        UNDEFINED_OFFSET,
        UNDEFINED_OFFSET,
        irBuiltIns.nothingNType
    )

fun irEmpty(context: JsIrBackendContext): IrExpression {
    return JsIrBuilder.buildComposite(context.dynamicType, emptyList())
}

fun IrSimpleFunction.isObjectInstanceGetter(): Boolean {
    return origin == JsLoweredDeclarationOrigin.OBJECT_GET_INSTANCE_FUNCTION
}

fun IrField.isObjectInstanceField(): Boolean {
    return origin == IrDeclarationOrigin.FIELD_FOR_OBJECT_INSTANCE
}

fun IrClass.findDefaultConstructorForReflection(): IrFunction? =
    defaultConstructorForReflection?.let { it.constructorFactory ?: it }

val IrClass.primaryConstructorReplacement: IrSimpleFunction?
    get() = findDeclaration<IrSimpleFunction> { it.isEs6PrimaryConstructorReplacement }

==> symbol-processing-aa-embeddable-2.2.0-2.0.2-sources.jar: org/jetbrains/kotlin/ir/backend/js/utils/IrJsUtils.kt <==
/*
 * Copyright 2010-2021 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.ir.backend.js.utils

import ksp.org.jetbrains.kotlin.ir.util.parents
import ksp.org.jetbrains.kotlin.descriptors.isClass
import ksp.org.jetbrains.kotlin.descriptors.isInterface
import ksp.org.jetbrains.kotlin.ir.UNDEFINED_OFFSET
import ksp.org.jetbrains.kotlin.ir.backend.js.JsIrBackendContext
import ksp.org.jetbrains.kotlin.ir.backend.js.JsLoweredDeclarationOrigin
import ksp.org.jetbrains.kotlin.ir.backend.js.JsStatementOrigins
import ksp.org.jetbrains.kotlin.ir.backend.js.constructorFactory
import ksp.org.jetbrains.kotlin.ir.backend.js.defaultConstructorForReflection
import ksp.org.jetbrains.kotlin.ir.backend.js.export.isExported
import ksp.org.jetbrains.kotlin.ir.backend.js.ir.JsIrBuilder
import ksp.org.jetbrains.kotlin.ir.backend.js.lower.*
import ksp.org.jetbrains.kotlin.ir.declarations.*
import ksp.org.jetbrains.kotlin.ir.expressions.IrExpression
import ksp.org.jetbrains.kotlin.ir.expressions.IrFunctionAccessExpression
import ksp.org.jetbrains.kotlin.ir.expressions.IrInlinedFunctionBlock
import ksp.org.jetbrains.kotlin.ir.expressions.IrReturn
import ksp.org.jetbrains.kotlin.ir.expressions.impl.IrConstImpl
import ksp.org.jetbrains.kotlin.ir.expressions.impl.IrGetFieldImpl
import ksp.org.jetbrains.kotlin.ir.symbols.IrFunctionSymbol
import ksp.org.jetbrains.kotlin.ir.symbols.IrReturnableBlockSymbol
import ksp.org.jetbrains.kotlin.ir.types.classifierOrNull
import ksp.org.jetbrains.kotlin.ir.util.*
import ksp.org.jetbrains.kotlin.name.FqName

val IrFile.nameWithoutExtension: String get() = name.substringBeforeLast(".kt")

fun IrClass.jsConstructorReference(context: JsIrBackendContext): IrExpression {
    return JsIrBuilder.buildCall(context.intrinsics.jsClass, origin = JsStatementOrigins.CLASS_REFERENCE)
        .apply {
            typeArguments[0] = defaultType
        }
}

fun IrDeclaration.isExportedMember(context: JsIrBackendContext) =
    parentClassOrNull != null && isExported(context)

fun IrDeclaration?.isExportedClass(context: JsIrBackendContext) =
    this is IrClass && kind.isClass && isExported(context)

fun IrDeclaration?.isExportedInterface(context: JsIrBackendContext) =
    this is IrClass && kind.isInterface && isExported(context)

fun IrReturn.isTheLastReturnStatementIn(target: IrReturnableBlockSymbol): Boolean {
    val ownerFirstStatement = target.owner.statements.singleOrNull()
    if (ownerFirstStatement is IrInlinedFunctionBlock) {
        return ownerFirstStatement.statements.lastOrNull() === this
    }
    return target.owner.statements.lastOrNull() === this
}

fun IrDeclarationWithName.getFqNameWithJsNameWhenAvailable(shouldIncludePackage: Boolean): FqName {
    val name = getJsNameOrKotlinName()
    return when (val parent = parent) {
        is IrDeclarationWithName -> parent.getFqNameWithJsNameWhenAvailable(shouldIncludePackage).child(name)
        is IrPackageFragment -> getKotlinOrJsQualifier(parent, shouldIncludePackage)?.child(name) ?: FqName(name.identifier)
        else -> FqName(name.identifier)
    }
}

fun IrConstructor.hasStrictSignature(context: JsIrBackendContext): Boolean {
    val primitives = with(context.irBuiltIns) { primitiveTypesToPrimitiveArrays.values + stringClass }
    return with(parentAsClass) {
        isExternal || isExpect || isAnnotationClass || context.inlineClassesUtils.isClassInlineLike(this) || symbol in primitives
    }
}

private fun getKotlinOrJsQualifier(parent: IrPackageFragment, shouldIncludePackage: Boolean): FqName? {
    return (parent as? IrFile)?.getJsQualifier()?.let { FqName(it) } ?: parent.packageFqName.takeIf { shouldIncludePackage }
}

val IrClass.isInstantiableEnum: Boolean
    get() = isEnumClass && !isExpect && !isEffectivelyExternal()

val IrDeclaration.parentEnumClassOrNull: IrClass?
    get() = parents.filterIsInstance<IrClass>().firstOrNull { it.isInstantiableEnum }

fun IrFunctionSymbol.isUnitInstanceFunction(context: JsIrBackendContext): Boolean {
    return owner.origin === JsLoweredDeclarationOrigin.OBJECT_GET_INSTANCE_FUNCTION &&
            owner.returnType.classifierOrNull === context.irBuiltIns.unitClass
}

// TODO: the code is written to pass Repl tests, so we should understand. why in Repl tests we don't have backingField
fun JsIrBackendContext.getVoid(): IrExpression =
    intrinsics.void.owner.backingField?.let {
        IrGetFieldImpl(
            UNDEFINED_OFFSET,
            UNDEFINED_OFFSET,
            it.symbol,
            irBuiltIns.nothingNType
        )
    } ?: IrConstImpl.constNull(
        UNDEFINED_OFFSET,
        UNDEFINED_OFFSET,
        irBuiltIns.nothingNType
    )

fun irEmpty(context: JsIrBackendContext): IrExpression {
    return JsIrBuilder.buildComposite(context.dynamicType, emptyList())
}

fun IrSimpleFunction.isObjectInstanceGetter(): Boolean {
    return origin == JsLoweredDeclarationOrigin.OBJECT_GET_INSTANCE_FUNCTION
}

fun IrField.isObjectInstanceField(): Boolean {
    return origin == IrDeclarationOrigin.FIELD_FOR_OBJECT_INSTANCE
}

fun IrClass.findDefaultConstructorForReflection(): IrFunction? =
    defaultConstructorForReflection?.let { it.constructorFactory ?: it }

val IrClass.primaryConstructorReplacement: IrSimpleFunction?
    get() = findDeclaration<IrSimpleFunction> { it.isEs6PrimaryConstructorReplacement }
